package kraft;

//
// a Prolog term represented as a String
//
// Author:	K.Y. Hui
// Date:	26 June 1998
// Version:	1.20
//		Now handle single-quotation in argument properly
//

public class TermInString
{
private final int TYPE_UNKNOWN=0;
private final int TYPE_ATOM=1;
private final int TYPE_LIST=2;
private final int TYPE_STRING=3;
private final int TYPE_EMPTY_LIST=4;
private final int TYPE_COMPOUND_TERM=5;

public String term;			//the term as a string
public String functor=null;
public int arity=0;
private int argument_start[];
private int argument_end[];
private int type=TYPE_UNKNOWN;

//
// constructor
// create a term and find arity
//
public TermInString(String s)
	throws Exception
	{
	int left=0;
	int right=s.length()-1;

	term=s;					//the term as a string
	if (s.charAt(left)=='"')		//looks like a string
		{
		left=skip_string(s,left);
		if (left<=right)		//still something left
			throw new Exception("malformed string");
		else	{
			functor=".";		//a string is a list with functor '.'
			arity=2;		//of arity 2
			type=TYPE_STRING;
			}
		}
	else if (s.charAt(left)=='[')		//looks like a list
		{
		left=skip_list(s,left);		//skip after a list
		if (left<=right)
			throw new Exception("malformed list");
		if (left==2)
			{
			functor="[]";		//an empty list is an atom
			arity=0;
			type=TYPE_EMPTY_LIST;
			}
		else	{
			functor=".";
			arity=2;
			type=TYPE_LIST;
			}
		}
	else	{
		left=skip_atom(s,left);			//skip functor

		if (left>right)				//end of string after skipping an atom
			{
			if (s.charAt(0)=='\'')
				functor=s.substring(1,s.length()-1);
			else functor=s;			//functor is the whole term
			arity=0;			//of arity 0
			type=TYPE_ATOM;
			}
		//
		// it must be a compound term
		//
		else	{				//now that's a compound term
			int arg_count=0;		//counting arguments
			int comma_count=0;		//counting commas
			int temp_start[];
			int temp_end[];

			//
			// create temp arrays for argument indexes
			//
			for (int i=0;i<s.length();i++)	//estimate the no. of argument by counting commas
				if (s.charAt(i)==',')
					comma_count++;
			temp_start=new int[comma_count+1];	//arg is 1 more than no. of commas
			temp_end=new int[comma_count+1];


			if (s.charAt(0)=='\'')			//if there are quotes at both ends
				functor=s.substring(1,left-1);
			else functor=s.substring(0,left);	//get the functor
			if (s.charAt(left)!='(')	//not a bracket following
				{
				throw new Exception("unexpected char following an atom");
				}

			//
			// now create indexes to arguments
			//
			do	{
				left++;					//skip after to next char
				temp_start[arg_count]=left;		//start of an argument
				left=skip_term(s,left);			//skip a term
				temp_end[arg_count]=left;		//end of an argument
				arg_count++;				//get ready for next argument
				arity++;				//1 more argument
				} while (s.charAt(left)==',');		//loop while there is a comma following
			if (s.charAt(left)!=')')
				throw new Exception("malformed compound term: missing ')'"+s.charAt(left));
			left++;
			if (left<=right)
				throw new Exception("malformed compound term");
			argument_start=new int[arg_count];		//create real indexing arrays
			argument_end=new int[arg_count];
			for (int i=0;i<arg_count;i++)
				{
				argument_start[i]=temp_start[i];
				argument_end[i]=temp_end[i];
				}
			type=TYPE_COMPOUND_TERM;
			}
		}
	}

public boolean is_atom()
	{
	return(type==TYPE_ATOM);
	}

public boolean is_empty_list()
	{
	return(type==TYPE_EMPTY_LIST);
	}

public boolean is_list()
	{
	return(type==TYPE_EMPTY_LIST || type==TYPE_LIST);
	}

public boolean is_string()
	{
	return(type==TYPE_STRING);
	}

public boolean is_compound_term()
	{
	return(type==TYPE_COMPOUND_TERM);
	}

//
// find a particular argument of a term
//
public String argument(int pos)
	{
	int left=0;
	int count=0;

	if (pos>arity)			//out of range
		return(null);
	switch (type)
		{
		case TYPE_COMPOUND_TERM:	return(term.substring(argument_start[pos-1],argument_end[pos-1]));
		default:			return(null);
		}
	}


//
// skip after an atom
//
private int skip_atom(String s,int left)
	throws StringIndexOutOfBoundsException, Exception
	{
	char c;
	boolean single_quote=false;

	c=s.charAt(left);
	if (c=='\'')			//single-quote at the beginning
		{
		single_quote=true;
		left++;
		}
	while (left<s.length())
		{
		c=s.charAt(left);
		switch (c)
			{
			case '\\':	left++;			//skip next char
					break;
			case '\'':	if (!single_quote)	//no single-quote at the beginning
						throw new Exception("unexpected single-quote in an atom");
					return(left+1);		//otherwise it is end of atom
			case '"':	if (!single_quote)
						throw new Exception("unexpected double-quote in an atom");
					break;
			case '[':
			case ']':
			case '(':
			case ',':
			case ')':	if (!single_quote)	//if not inside a single quote
						return(left);	//then we hit the end of an atom
					break;
			default:	break;
			}
		left++;
		}
	return(left);
	}

//
// skip after a term
//
private int skip_term(String s,int left)
	throws StringIndexOutOfBoundsException, Exception
	{
	char c;

	c=s.charAt(left);
	if (c=='"')
		return(skip_string(s,left));
//	if (c=='\'')
//		return(skip_atom(s,left));
	if (c=='[')
		return(skip_list(s,left));
	try	{
		left=skip_compound_term(s,left);	//try a compound term 1st
		return(left);
		} catch (Exception e)
			{
			}
	return(skip_atom(s,left));			//then try atom
	}

//
// return index just after a string
//
private int skip_string(String s,int left)
	throws StringIndexOutOfBoundsException, Exception
	{
	char c;

	if (s.charAt(left)!='"')	//check for leading double-quote
		throw new Exception("missing double-quote in the beginning of a string");

	left++;			//skip double-quote
	try	{
		c=s.charAt(left);
		while (c!='"')	//loop until hitting quote
			{
			switch (c)
				{
				case '\\':	left++;		//skip char after escape
						break;
				default:	break;
				}
			left++;					//next char
			c=s.charAt(left);
			}
		return(left+1);					//double-quote is found
		} catch (StringIndexOutOfBoundsException e)
			{
			throw new Exception("missing quote at end of string");
			}
	}

private int skip_list(String s,int left)
	throws StringIndexOutOfBoundsException, Exception
	{
	if (s.charAt(left)!='[')
		throw new Exception("missing square-bracket at the beginning of a list");
	left++;						//skip square bracket
	try	{
		if (s.charAt(left)==']')		//check for empty list
			return(left+1);
		left=skip_argument_list(s,left);	//skip a list of arguments
			if (s.charAt(left)==']')	//look for matching ']'
				return(left+1);		//return index just after ']'
			else throw new Exception("missing square-bracket at the end of a list");
		} catch (StringIndexOutOfBoundsException e)
			{
			throw new Exception("missing square-bracket at the end of a list");
			}
	}

//
// skip after a list of arguments
//
private int skip_argument_list(String s,int left)
	throws StringIndexOutOfBoundsException, Exception
	{
	
	try	{
		left=skip_term(s,left);			//skip after term
		if (s.charAt(left)!=',')
			return(left);
		do	{
			left++;
			left=skip_term(s,left);			//skip after term
			} while (s.charAt(left)==',');		//loop while there is a comma following
		return(left);
		} catch (StringIndexOutOfBoundsException e)
			{
			return(left);
			}
	}

//
// skip after a compound term
//
private int skip_compound_term(String s, int left)
	throws StringIndexOutOfBoundsException, Exception
	{
	left=skip_atom(s,left);					//skip functor
	if (s.charAt(left)!='(')				//look for starting bracket
		throw new Exception("missing bracket at the beginning of a compound term");
	left++;
	left=skip_argument_list(s,left);			//skip list of arguments
	if (s.charAt(left)!=')')
		throw new Exception("missing bracket at the end of a compound term");
	return(left+1);
	}
}

