package kraft.monitor.adapter;


import kraft.monitor.display.Message;
import kraft.monitor.tools.KraftTerm;

import kraft.TermInString;



/** Represents a CCQL message as an object.
  * This means this message can be passed to the Monitor
  * like any other Message object
  *
  * At the moment this is used in conjunction with Kraft_msg,
  * to format the CCQL part of these messages in a way
  * suitable for visualisation.
  * 
  * Strings representing prolog term structures (ccql/2) are
  * parsed by this class.
  *
  *@author Ted Francis
  *@see Kraft_msg
  */
//====================================
public class CCQL implements Message {
//====================================


  public String performative = "", 
                sender       = "", 
                receiver     = "",
                language     = "",
                content      = "",
                reply_with   = "",
                in_reply_to  = "";

  private String the_term = null;
  private String formatted= null;



  /** Create a new CCQL message with the given string
    *
    *@exception Exception if the string is not a valid ccql term
    */
  //----------------------------------------
  public CCQL(String str) throws Exception {
  //----------------------------------------
    this(new TermInString(str));
  }

  /** Create a new CCQL message with the given Term
    *
    *@exception Exception if the string is not a valid ccql term
    */
  //-----------------------------------------------
  public CCQL(TermInString term) throws Exception {
  //-----------------------------------------------

    //make sure it is a valid term structure
    if(term.functor.equals("ccql") && term.arity == 2) {
      TermInString list = new TermInString(term.argument(2));

      if(list.is_empty_list() || 
	 (list.functor.equals(".") && list.arity ==2)) {
	
	performative = term.argument(1);
	sender       = KraftTerm.getFieldValue(list,"sender");
	receiver     = KraftTerm.getFieldValue(list,"receiver");
	content      = KraftTerm.getFieldValue(list,"content");
	language     = KraftTerm.getFieldValue(list,"language");
	reply_with   = KraftTerm.getFieldValue(list,"reply_with");
	in_reply_to  = KraftTerm.getFieldValue(list,"in_reply_to");
	
	this.the_term = term.term;
      }
    }
    if(the_term==null) 
      throw new Exception("Not valid ccql term: \""+term+"\"");
    
     System.out.println("Parsed CCQL term OK --> "+
		       "\n\tFULL         = "+the_term+
		       "\n\tperformative = "+performative+
		       "\n\tsender       = "+sender+
		       "\n\treceiver     = "+receiver+
		       "\n\tlanguage     = "+language+
		       "\n\tcontent      = "+content);
 

  }//end constructor



  /** Parse the content field of the message.  
    * Checks to see whether the string is a term, 
    * and if so, formats it.
    */
  protected String formatContent(String content, int level){
    TermInString term;
    String rtn = "";

    try{
      term = new TermInString(content);


      // If the content is another ccql message
      //---------------------------------------
      if(term.functor.equals("ccql") && term.arity==2) {

	//Add tabs to the Term's display string so it is indented more
	// than the current level
	StringBuffer buf = 
	  new StringBuffer((new CCQL(term)).getDisplayString(1));

	int i=0;
	if(level<=1) buf.insert(i++,'\n');
	buf.insert(i,tabs(level+1));
	for(i=i; i<buf.length(); i++){
	  if(buf.charAt(i) == '\n') 
	    buf.insert(i+1, tabs(level+1));
	}//end for
	rtn = buf.toString();
      }

      // If the content is a list of terms -- recurse on each element
      //-------------------------------------------------------------
      else if(term.functor.equals(".") && term.arity==2){ 

	if(level>1) rtn = tabs(level+1) + "[";
	else        rtn = "[";
	boolean at_start = true;
	while( ! (term==null || 
		  term.is_empty_list() || 
		  term.argument(1)==null ||
		  term.argument(2)==null))
	  {
	    if( ! at_start)rtn += ",";
	    rtn += "\n";
	    //rtn += tabs(level+1);
	    rtn += formatContent(term.argument(1),level+1);
	    term = new TermInString(term.argument(2));

	    // If the list does not continue with more list
	    if( term.functor.equals(".") == false) {
	      if( term.is_empty_list() ==false)
		rtn +=(",\n"+tabs(level+2)+term.term);
	      break;//from while loop
	      }
	    at_start = false;
	  }//while
	if(level>1) rtn += "\n" + tabs(level+1) + "]";
	else        rtn += "\n" + tabs(level)   + "]";
      }

      
      // If the content is some other term
      //----------------------------------
      else {
	if(level>1) rtn = tabs(level+1) + term.term;
	else rtn = term.term;
      }
    }
    catch(Exception e) { 
      // If anything goes wrong return original content
      // this WILL happen if the content is not a term, this is correct
      rtn = content; 
    }

    return rtn;
  }//end formatContent



  private String tabs(int n){
    String r ="";
    for(int i=1;i<=n;i++)
      r += "     ";//5 spaces per tab
    return r;
  }

  /** Returns a formatted string representing the CCQL
    * message.  In this implementation, the value of detail
    * is ignored, the same string is always returned
    *
    * Needed so this class can implement the Message interface
    *
    *@see kraft.monitor.display.Message
    */
  //----------------------------------------
  public String getDisplayString(int detail){
  //----------------------------------------
    
    StringBuffer buf;
    TermInString list, colon_part;
    boolean at_start;

    if(formatted==null) {
      buf = new StringBuffer(20);
      try{
	TermInString my_term = new TermInString(the_term);
	buf.append(my_term.functor);
	buf.append("(");
	buf.append(my_term.argument(1));
	buf.append(",");
	
	list = new TermInString(my_term.argument(2));
	
	if(list.is_empty_list())
	  buf.append(" [])");

	else if(list.functor.equals(".") && list.arity==2) {
	  buf.append(" [");
	  at_start = true;

	  while( ! (list==null ||
		    list.is_empty_list() ||
		    list.argument(1)==null ||
		    list.argument(2)==null )) 
	    {
	      colon_part = new TermInString(list.argument(1));
	      if(at_start==false) {
		buf.append(",");
	      }
	      buf.append("\n");
	      buf.append(tabs(1));
	      
	      if(colon_part.functor.equals(":") && colon_part.arity==2) {
		buf.append(colon_part.argument(1));
		buf.append(" : ");
		if(colon_part.argument(1).equals("content"))
		  buf.append(formatContent(colon_part.argument(2), 1));
		else
		  buf.append(colon_part.argument(2));
	      }
	      else {
		buf.append(colon_part.term);//just put it on the buffer
	      }
	      at_start=false;

	      //goto next list element
	      list  = new TermInString(list.argument(2));
	      
	      //if the list does not continue with 
	      if( list.functor.equals(".") == false) {
		if( list.is_empty_list() ==false)
		  buf.append(",\n"+tabs(1)+list.term);
		break; //from while loop
	      }

	    }//end while
	  buf.append("\n])");
	}//end else if
      }
      catch(Exception e){
	System.err.println("Error parsing args in list.");
      }
      formatted = buf.toString();
    }
    return formatted;
  }//end getDetailString();



  /** Returns the performative of the CCQL message.
    * Needed so this class can implement the Message interface
    */
  public String getCommandString(){
    return performative;
  }


  /** Returns the original, unformatted term structure.
    * Needed to implement the Message interface
    */
  public String toString(){
    return the_term;
  }

  /** Whether another object represents the same message 
    * as this one.
    * Needed so this class can implement the Message interface
    */
  public boolean equals(Object obj) {
    if (obj instanceof CCQL && 
	obj.toString().equals(this.toString()))
      return true;
    else
      return false;
  }
}//end class CCQL
