package kraft.monitor.adapter;

import kraft.monitor.display.Monitorable;
import kraft.TermInString;

import java.util.*;

/** Implements the requirements for a Monitorable object.  
  * Used to represent a Kraft Agent so it can be registered by the
  * Monitor
  *<br>
  * Class could be extended to allow the Monitor to 'callback' this
  * object, which would act as a proxy for the actual (possibly Prolog)
  * agent, communicate the monitor's request to the agent, and receive
  * replies from the agent 
  *
  *@author Ted Francis
  */
//==============================================
public class  KraftAgent implements Monitorable{
//==============================================

  private String full_name;
  private int    type;

  private String name     = null, 
                 site     = null,
                 role     = "",
                 resource = "",
                 protocol = "";

  private Hashtable infoPairs;


  /** Create a KraftAgent with the given name
    *
    */
  KraftAgent(String n) throws NullPointerException {
    if(n==null) throw new NullPointerException("kraftAgent creation");

    full_name = n;
    try{
      TermInString term = new TermInString(n);
      if(term.functor.equals("krl") && term.arity == 2) {
	site = term.argument(1);
	name = term.argument(2);
      }
    }catch(Exception e) {}
    
    if(name==null) name = n;

    infoPairs = new Hashtable(2);
    type = MIDDLE;

  }


  /** Gets the name of this agent in a formatted form.
    *@param detail_level if 1, returns just the name of the agent
    *                    if 2, returns more detail about the agent
    */
  public String getDisplayName(int detail_level) {

    Enumeration e;
    String str = full_name; //default value

    if(detail_level<=0) detail_level = 1;
    if(detail_level>2)  detail_level = 2;

    switch(detail_level){
    case 1:  if(site==null) str = name;
             else str = (site + "/" + name);
             break;
    case 2:  if(site==null) 
               str = ("name: " + full_name);
             else
	       str = ("name: " + name + "\nsite: " + site);

             if(role.length()>0)
	       str = (str + "\nrole: " + role);
             if(resource.length()>0)
	       str = (str + "\nresource: " + resource);
	     if(protocol.length()>0)
	       str = (str + "\nprotocol: " + protocol);

	     if( ! infoPairs.isEmpty()) {
	       e = infoPairs.keys();
	       while(e.hasMoreElements()){
		 String s = (String)e.nextElement();
		 if(s.equals("")) break;//from while loop
		 str = (str + "\n" + s + ": " + infoPairs.get(s));
	       }
	     }
             break;
    }
    return str;
  }

  
  /** The unformatted name of the agent
   *
   */
  public String getFullName() {
    return full_name;
  }

  /**The Monitorable type of the agent. 
    *eg. BROKER, USER, MIDDLE, SERVER
    */
  public int getType() {
    return type;
  }

  /** The name of the resource this agent 
    * wraps onto (if any). Otherwise returns ""
    */
  public String getResourceType() {
    return resource;
  }

  /** The protocol used between the agent and its
    * resource (if any). Otherwise returns ""
    */
  public String getResourceProtocol() {
    return protocol;
  }
  

  /** Set the KRAFT role of this agent. Used to discern the Monitorable
    * type of the agent for display purposes. The agent will be displayed
    * on the screen in different areas depending on the role
    *
    *@see #getType()
    *
    *@param role Any one of "ontology_agent", "ontologyagent", "facilitator", 
    *                       "user_agent",     "useragent",     "auxiliary",
    *                       "mediator",       "resource",      "wrapper" 
    */
  public void setRole(String role){
    this.role = role;
    
    if (role.equalsIgnoreCase("ontology_agent") ||
	role.equalsIgnoreCase("ontologyagent")  ||
	role.equalsIgnoreCase("facilitator")      )
	
        type = BROKER;
    else
    if (role.equalsIgnoreCase("user_agent")     ||
	role.equalsIgnoreCase("useragent")        )

	type = USER;
    else
    if (role.equalsIgnoreCase("auxiliary")      ||
	role.equalsIgnoreCase("mediator")         )

        type = MIDDLE;
    else 
    if (role.equalsIgnoreCase("resource")       || 
	role.equalsIgnoreCase("wrapper")          )

        type = SERVER;
    else
      type = MIDDLE; //default

  }//end setRole

  

  /** Sets the type of resource, and protocol this agent uses
    * to communicate with that resource
    *
    * Set either param to null or "" if required.
    */
  public void setResource(String resource, String protocol) {

    //only use this if the agent is a wrapper for a resource
    if(type==SERVER) {
      this.resource = resource;
      this.protocol = protocol;
    }
  }


  /** Adds more info to the agent, in a similar way to
    * resource: type, protocol: type,
    * This info is used when getDisplayString(2) is 
    * called.
    *
    *@param field any label 
    *@param value any label --  the info pair will be shown thus 
    *                           ( field : value )
    */
  public void addInfoPair(String field, String value) {
    infoPairs.put(field,value);
    System.out.println(getFullName()+": info pair added :("+field+","+value+")");
  }

}//end class KraftAgent
















