package kraft.monitor.display;

import kraft.monitor.tools.Debug;

import java.awt.*;
import java.util.Vector;
import java.util.Enumeration;


/** Defines a rectangular area which contains a number of Nodes.    <br><br>
  * The MonitorCanvas will own several of these objects, each
  * to represent different areas of the canvas.                     <br><br>
  * An Area has a type associated with it, corresonding to the type
  * of 'agents' it displays.  All Nodes in the Area will have this
  * same type (any of the types in the Monitorable interface).      <br><br>
  *
  * The Area can lay out the Nodes it contains in different ways:
  * a horizontal line, a vertical line, or a circle.  It keeps the
  * Nodes correctly laid out as they are added to and removed from the Area. <br>
  * The area initialises with a default layout based on its type, but 
  * this can be changed using setNodeLayout.
  *
  *@author Ted Francis
  *@see    #setNodeLayout
  *@see    kraft.monitor.display.Node
  *@see    kraft.monitor.display.MonitorCanvas
  *@see    kraft.monitor.display.Monitorable
  */

//================================
public class Area extends Rectangle {
//================================

  /** The minimum amount of space between adjacent Nodes in this Area and 
    * the position of the nearest Node (defaults to 15 pixels) */
  public static int        V_SPACE     = 15, H_SPACE   = 15;

  /** Constant used to adjust the Node layout rule used.
    *@see #setNodeLayout
    */
  public static final int HORIZ_LINE  = 0,  VERT_LINE = 1, CIRCLE = 2;


  /** Used when the Node layout is CIRCLE, to determine where the first Node in circle 
    * is positioned*/
  private static final double    START_ANGLE = 3 * Math.PI/2;  //12 o'clock position


  private int      nodeLayout       = VERT_LINE;  
  private Vector   nodes;
  private int      type;
  private String   typeStr;
  private int      labelOrientation = Node.EAST;
  private boolean  messy            = true;
  private Color    background       = Color.white;

  // Constructor methods
  // ===================


  /** Start the new Area with the given type of Nodes it should display
    * Type should be any type in the Monitorable interface.
    *@see kraft.monitor.display.Monitorable
    */
  //----------------
  public Area(int t){
  //----------------
    super();
    setUp(t);
  }

  /** Start the new Area with the given type, position, and dimensions.
    * Type should be any type in the Monitorable interface.
    *@see Monitorable
    */
  //------------------------------------------
  public Area(int t, int x, int y, int w, int h){
  //------------------------------------------
    super(x,y,w,h);
    setUp(t);
  }


  /** Work out what default layout to use based on the type.
    * The owner of this Area can change this using 
    * setNodeLayout if required
    */
  //-------------------------
  private void setUp(int t){
  //-------------------------
    type = t;
    nodes = new Vector(10, 5);

    switch(type){
      case Monitorable.BROKER: typeStr = "BROKER"; 
	                       setNodeLayout(HORIZ_LINE);
			       setLabelOrientation(Node.SOUTH);
			       break;

      case Monitorable.USER:   typeStr = "USER";   
	                       setNodeLayout(VERT_LINE);
			       setLabelOrientation(Node.WEST);
			       break;

      case Monitorable.MIDDLE: typeStr = "MIDDLE"; 
	                       setNodeLayout(CIRCLE);
			       setLabelOrientation(Node.SOUTH);
			       break;

      case Monitorable.SERVER: typeStr = "SERVER"; 
	                       setNodeLayout(VERT_LINE);
			       setLabelOrientation(Node.EAST);
			       break;
    }
  }//end setUp



  public boolean hasNodes() {
    return (nodes.size() > 0);
  }



  /** Adds a Node to this Area with the orientation specified. <br>
    * Adjusts position of all Nodes to make space for it.
    *
    *@param b      The Node to add
    *@param orient Any of the compass point values defined in Node
    *@see #setLabelOrientation
    */
  //----------------------------------
  public void add(Node b, int orient){
  //----------------------------------
    nodes.addElement(b);
    b.setLabelOrientation(orient);
    calculatePositions();
    messy = true; 
 }


  /** Adds a Node to this Area, it applies the <i> default labelOrientation</i>
    * to the Node for this type of Area. <br>
    * Adjusts position of all Nodes to make space for it.
    *@see #setLabelOrientation
    */
  //-----------------------
  public void add(Node b) {
  //-----------------------
    nodes.addElement(b);
    b.setLabelOrientation(labelOrientation);
    Debug.println("Area("+typeStr+"): node added, now have "+nodes.size());
    calculatePositions();
    messy = true; 

  }//end add(Node)


  /** Remove the Node from this area and adjust positions of remaining
    * Nodes to compensate */
  //------------------------
  public void remove(Node b) {
  //------------------------
    nodes.removeElement(b);
    calculatePositions();
    Debug.println("Area("+typeStr+"): node removed, now have "+nodes.size());
    messy = true; 
  }


  /** Calculate positions for all Nodes based on the layout style */
  //------------------------------
  private void calculatePositions() {
  //------------------------------
    int stepX, stepY, nodeX, nodeY;
    stepX = stepY = nodeX = nodeY = 0;
    int   radius = 0;
    double theta = 0;
    

    if(width>0 && height>0) {

    switch(nodeLayout) {
       case HORIZ_LINE: stepX = width / (nodes.size() + 1);
	                stepY = 0;
	                nodeX = x + stepX;
			nodeY = y + height/2;
	                break;
        case CIRCLE:    radius = Math.min(width/3, height/3);
	                theta  = START_ANGLE; 
			nodeX  = x + width/2;
			nodeY  = y + height/2;
			break;
        case VERT_LINE: stepX = 0;
	                stepY = height / (nodes.size() + 1);
	                nodeX = x + width/2;
			nodeY = y + stepY;
			break;

    }
    //int xoffset = 0, yoffset = 0;

    Enumeration e = nodes.elements();
    int i=0;
    while(e.hasMoreElements()) {
      Node node = (Node)e.nextElement();

      if(nodeLayout==CIRCLE && nodes.size() > 1) {
	nodeX  =  x + width/2  + (int) (radius * Math.cos(theta));
	nodeY  =  y + height/2 + (int) (radius * Math.sin(theta));
	theta += (2 * Math.PI)/nodes.size();
      }

      node.setPosition(nodeX, nodeY);
      //Rectangle br = node.getBounds();

      if(nodeLayout != CIRCLE) {
	nodeX += stepX;
	nodeY += stepY;
      }
    }//end while      
  
    messy = true;
    }
  }//end calculatePositions


  /** Completly wipe the contents of this area from the graphics context */
  //------------------------------
  private void unpaint(Graphics g) {
  //------------------------------
    //System.out.println("Area["+typeStr+"].UNpaint: entered");
    g.setColor(getBackground());
    g.fillRect(x,y,width,height);
  }


  /** Paint the nodes in this area onto the given graphics context 
   *@see kraft.monitor.display.Node#paint
   */
  //---------------------------
  public void paint(Graphics g) {
  //---------------------------
    
    //    System.out.println("Area["+typeStr+"].paint: entered");

    if (messy) {
      unpaint(g);
      messy = false;
    }
    if(width>0 && height > 0) {
      Enumeration e = nodes.elements();
      while(e.hasMoreElements()) {
	Node b = (Node) e.nextElement();
	b.paint(g);
      }
    }

  }//end paint


  /** Set the maximum size for all the Nodes in this Area
    */
  //--------------------------------------
  public void setContainerSize(Dimension d) {
  //--------------------------------------
    Enumeration e = nodes.elements();
    while(e.hasMoreElements()) {
      ((Node)e.nextElement()).setContainerSize(d);
    }
  }

  private Graphics myGraphics;


  /** Set the graphics context for drawing onto*/
  //---------------------------------
  public void setGraphics(Graphics g) {
  //---------------------------------

    myGraphics = g;
  }


  private Node selectedNode=null;


  /** Determine whether a mouse press at (x,y) has 
    * an effect on any of the Nodes in this Area.
    * If so, make that effect occur, and return true, else return false
    *@see kraft.monitor.display.Node#mousePressedHasEffect
    */
  //----------------------------------------------
  public boolean mousePressedHasEffect(int x, int y) {
  //----------------------------------------------
    if(this.contains(x,y)) {
      Enumeration e = nodes.elements();
      while(e.hasMoreElements()) {
	Node b = (Node)e.nextElement();

	if(b.mousePressedHasEffect(x,y)) {
	  selectedNode = b;
	  selectedNode.paint(myGraphics); //makes sure it is painted on TOP
	  return true;
	}
      }//while
    }//if

    return false;
  }//end mousePressedHasEffect
      



  /** Determine whether a mouse released at (x,y) has 
    * an effect on any of the Nodes in this Area.
    * If so, make that effect occur, and return true, else return false
    *@see kraft.monitor.display.Node#mouseReleasedHasEffect
    */
  //-----------------------------------------------
  public boolean mouseReleasedHasEffect(int x, int y) {
  //-----------------------------------------------
    
    if(selectedNode != null) {
      if(selectedNode.mouseReleasedHasEffect(x, y)) {
	selectedNode.unpaint(myGraphics);
	selectedNode.paint(myGraphics);
	selectedNode = null;
	return true;
      }
    }
    return false;
  }//end mousePressed



  //Mutator methods

  //--------------------------------
  public void setBackground(Color c) {
  //--------------------------------
    background = c;
  }
  

  /** Set the layout rule for the Nodes in this Area.
    *@see #HORIZ_LINE
    *@see #VERT_LINE
    *@see #CIRCLE
    */
  //-----------------------------
  public void setNodeLayout(int l) {
  //-----------------------------
    nodeLayout = l;
  }


  /** Set the default label orientation for all newly added Nodes
    * in this Area.
    *@see kraft.monitor.display.Node#NORTH
    */
  //-----------------------------------
  public void setLabelOrientation(int l) {
  //-----------------------------------
    labelOrientation = l;
  }


  /** Adjust the dimensions of this area by the given scaling factor.*/
  //-------------------------------------------------
  public void adjustBounds(double scaleX, double scaleY) {
  //-------------------------------------------------

    if (x>0) x = (int) Math.ceil( x * scaleX);
    if (y>0) y = (int) Math.ceil( y * scaleY);

    width  = (int) Math.ceil(width  * scaleX);
    height = (int) Math.ceil(height * scaleY);
    
    calculatePositions();
    messy = true;

  }

  /*override super-classes methods to ensure node positions
    are recalculated correctly*/
 
  public void setBounds(Rectangle r){
    super.setBounds(r);
    calculatePositions();
  }
  public void setBounds(int x, int y, int w, int h){
    super.setBounds(x,y,w,h);
    calculatePositions();
  }
  public void setSize(Dimension d){
    super.setSize(d);
    calculatePositions();
  }
  public void setSize(int w, int h){
    super.setSize(w,h);
    calculatePositions();
  }
  public void setLocation(Point p){
    super.setLocation(p);
    calculatePositions();
  }
  public void setLocation(int x, int y){
    super.setLocation(x,y);
    calculatePositions();
  }


  //Accessor methods

  //--------------------------
  public Color getBackground(){
  //--------------------------
    return background;
  }

  //-------------------------------
  public Dimension getMinimumSize() {
  //-------------------------------
    int minX, minY, maxX, maxY;

    if(nodes.isEmpty()) 
      return new Dimension(0,0);

    Point t = ((Node) nodes.firstElement()).getPosition();
    minX = maxX = t.x;
    minY = maxY = t.y;

    Enumeration e = nodes.elements();
    while(e.hasMoreElements()) {
      Point p = ((Node) e.nextElement()).getPosition();
      if(p.x < minX) minX = p.x;
      else if (p.x > maxX) maxX = p.x;
      if(p.y < minY) minY = p.y;
      else if (p.y > maxY) maxY = p.y;
    }

    return new Dimension(maxX - minX + 2*H_SPACE, maxY - minY + 2*V_SPACE);
  }//end getMinimumSize



  //----------------------
  public String toString() {
  //----------------------

    StringBuffer strbuf = new StringBuffer();

    strbuf.append("Area: type = "+typeStr+", pos = (" +x+ ","+y+")");
    strbuf.append(", dims = " + width + " x " + height);

    return strbuf.toString();
  }//end toString

}//end class Area














