package kraft.monitor.display;

import kraft.monitor.tools.Debug;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Graphics;
import java.awt.Polygon;
import java.awt.Color;
import java.awt.FontMetrics;


/** Defines the graphic representation of a Message passed between two Monitorable objects.
  *<br><br>
  * This is a 'link' between two Node objects.
  * Represented as an arrow containing the command name of the message.
  * A mouse click on the object ( via calls to <code> mousePressedHasEffect and mouseReleasedHasEffect </code>)
  * will cause this object to display more information about itself graphically when it is repainted.
  *@author Ted Francis
  *@see kraft.monitor.display.Message
  */

//=========================================
public class Link implements DisplayEvent {
//=========================================

  public static int  HSPACE = 5;

  private Node           startNode, endNode;
  private Point          start, end;
  private int            minX, minY, width, height;
  private String         label;
  private Message        message;
  private Date           time;
  private Color          background     = Color.white, 
                         foreground     = Color.blue;
  private Dimension      parentSize;
  private Rectangle      bounds, 
                         labelRect      = null;
  private boolean        mousePressed   = false, 
                         crossedOut     = false;
  private DisplayMonitor displayMonitor = null;


  /** A new link between node 'start' and node 'end'. The Message passed is m.
    * t is the time this event ocurred.
    */
  //----------------------------------------------------
  public Link(Node start, Node end, Message m, Date t) {
  //----------------------------------------------------
    startNode = start;
    endNode   = end;
    message   = m;
    time      = t;
    label     = m.getCommandString();
    updatePositions();
  }

  /** Unpaint the object from the graphics context */
  //------------------------------
  public void unpaint(Graphics gc) {
  //------------------------------
    int x,y,w,h;
    //System.out.println("Link["+label+"].UNpaint entered");

    x = bounds.x;
    y = bounds.y;
    w = bounds.width;
    h = bounds.height;

    gc.setColor(background);
    gc.fillRect(x,y,w,h);
  }


  /** Paint the object onto the graphics context */
  //----------------------------
  public void paint(Graphics gc) {
  //----------------------------
    boolean LtoR;
    String str;
    Point  mypoints[], newpoints[];


    //System.out.println("Link["+label+"].paint entered");

    updatePositions();

    //Debug.println("\nPAINT-----------------");
    //Debug.println("start = "+start+"\nend = "+end);
      
      // length of line I want by Pythag (a2 = b2 + c2)
      int length = (int) (Math.sqrt(Math.pow(end.x-start.x, 2) + 
				    Math.pow(end.y-start.y, 2)));

      //Debug.println("length is: "+length);
      
      //Set points for the arrow -- start pt based at(0,0)
      
      mypoints  = new Point[8]; //easy to work out points
      newpoints = new Point[mypoints.length]; //re-calculated points
      
      //points for an arrow-headed line
      mypoints[0] = new Point(0, 2);
      mypoints[1] = new Point(length-10, 2);
      mypoints[2] = new Point(length-10, 6);
      mypoints[3] = new Point(length,    0);
      mypoints[4] = new Point(length-10,-6);
      mypoints[5] = new Point(length-10,-1);
      mypoints[6] = new Point(0,-1);
      
      //the midpoint of the line to centre the label on
      mypoints[7] = new Point(length/2, 0);


      // Translate the points to be centered on the actual 'start' 
      // position, and rotate to lie along the line start->end
      //-------------------------------------------------------------
      rotateAndTranslate(start, end, start, mypoints, newpoints);
      // adjusted points appear in 'newpoints' array
      

      //make a polygon out of the points, and draw it
      Polygon arrow = new Polygon();
      for(int i = 0; i <= 6; i++) 
	{
	  arrow.addPoint(newpoints[i].x, newpoints[i].y);
	}
      gc.setColor(getForeground());
      gc.fillPolygon(arrow);
      
      //draw the label in a nice way
      FontMetrics metrics = gc.getFontMetrics();
      int labelW          = metrics.stringWidth(label) + HSPACE*2;
      int labelH          = metrics.getHeight();
      Point centre        = newpoints[7];
      
      labelRect = new Rectangle(centre.x - labelW/2, centre.y - labelH/2,
				labelW, labelH);

      //draw an empty box in centre of the line--print label in it
      gc.setColor(getBackground());
      gc.fillRect(labelRect.x, labelRect.y, labelRect.width, labelRect.height);
      gc.setColor(getForeground());
      gc.drawRect(labelRect.x, labelRect.y, labelRect.width, labelRect.height);
      
      gc.drawString(label,
		    centre.x - labelW/2 + HSPACE,
		    centre.y - labelH/2 + metrics.getAscent() + 1);
      
      if(crossedOut) {
	drawCrossOut(gc, labelRect);
      }

      setBounds(labelRect.union(arrow.getBounds()));

      if(mousePressed) {
	drawTooltip(gc, centre);
      }

      displayMonitor.setDisplayed(true);
  }//end paint



  /** Draw a red cross through the specified rectangle..
    * Will draw if setCrossedOut(true) has been called
    */
  private void drawCrossOut(Graphics gc, Rectangle rect) {
    
    Point   centre = new Point(rect.x + rect.width/2, rect.y + rect.height/2);
    Polygon cross  = new Polygon();
    
    cross.addPoint( rect.x,                  rect.y - 4               );
    cross.addPoint( rect.x + 1,              rect.y - 5               );
    cross.addPoint( centre.x,                centre.y - 1             );
    cross.addPoint( rect.x + rect.width - 1, rect.y - 5               );
    cross.addPoint( rect.x + rect.width,     rect.y - 4               );
    cross.addPoint( centre.x + 1,            centre.y                 );
    cross.addPoint( rect.x + rect.width,     rect.y + rect.height + 4 );
    cross.addPoint( rect.x + rect.width - 1, rect.y + rect.height + 5 );
    cross.addPoint( centre.x,                centre.y + 1             );
    cross.addPoint( rect.x + 1,              rect.y + rect.height + 5 );
    cross.addPoint( rect.x,                  rect.y + rect.height + 4 );
    cross.addPoint( centre.x - 1,            centre.y                 );


    gc.setColor(Color.red);
    gc.fillPolygon(cross);

  }//end drawCrossOut



  /** Draws the 'tooltip' for the Message object that this represents.
    * Uses the getFormattedString method from Message to do this.
    * Divides up the string based on the newline chars found, and then
    * splits it up more if the width exceeds the 'parentSize'
    */
  private void drawTooltip(Graphics gc, Point centre) {

    int x, y, w, h, maxW, maxH;
    FontMetrics fm;
    Vector strings;
    StringTokenizer tok;

    String original_string;
    original_string = replaceTabs(message.getDisplayString(1), 4);

      fm   = gc.getFontMetrics(); 
      tok  = new StringTokenizer(original_string, "\n\r");
      
      maxH = parentSize.height - 1;
      maxW = parentSize.width  - 1;
      
      strings = new Vector(tok.countTokens());
      
      w = 0;
      while (tok.hasMoreTokens())
      {
	String curr = tok.nextToken();
	int    strW = fm.stringWidth(curr);
	
	if(strW + HSPACE*2 <= maxW) 
	{
	  strings.addElement(curr);
	  w = Math.max(w, strW + HSPACE*2);
	}
	else
	{ //need to chop up the string into 'num' parts
	  int num = (int)Math.ceil((double)strW/ (maxW-HSPACE*2));  
	  //each part will be 'len' chars long
	  int len = curr.length()/num;
	  
	  for(int pos = 0; pos < curr.length() ; pos+=len)
	  {
	    int end; String tmp;
	    end = (pos+len>=curr.length()) ? curr.length()-1:pos+len;
	    tmp = curr.substring(pos, end);
	    strings.addElement(tmp);
	    w   = Math.max(w, fm.stringWidth(tmp)+HSPACE*2);
	  }
	}//end else

      }//end for

      
      h = strings.size() * fm.getHeight();
      if(crossedOut) {
	h += fm.getHeight(); //room for the extra 'not delivered' line
      }
      
      x = Math.max(0, centre.x - w/2);
      y = Math.max(0, centre.y - h/2);

      if(y+h > maxH) { 
	y = Math.max(0, y - (y+h-maxH));
      }
      if(x+w > maxW) {
	x = Math.max(0, x - (x+w-maxW));
      }

      //   gc.setColor(Color.cyan);
      //   gc.drawString("("+x+","+y+")  "+w+" x "+h, centre.x, centre.y - h);

      float[] hsbvals = new float[3];
      Color.RGBtoHSB(255,255,100, hsbvals);
      //a pale yellow
      gc.setColor(Color.getHSBColor(hsbvals[0],hsbvals[1],hsbvals[2])); 
      gc.fillRect(x,y,w,h);
      gc.setColor(Color.darkGray);
      gc.drawRect(x,y,w,h);
      
      x += HSPACE; 
      y += fm.getMaxAscent() + 1;

      if(crossedOut) {
	gc.setColor(Color.red);
	gc.drawString("MESSAGE NOT DELIVERED", x, y);
	y += fm.getHeight();
      }
      
      gc.setColor(Color.black);

      Enumeration e = strings.elements();
      while(e.hasMoreElements()) {
	gc.drawString((String)e.nextElement(), x, y);
	y += fm.getHeight();
      }


  }//end drawTooltip



  private String replaceTabs(String orig, int num_spaces){
    if(orig.indexOf('\t')>=0) { //there is a tab
      String spaces ="";
      int i;
      for(i=1;i<num_spaces;i++)
	spaces += " ";
      StringBuffer buf = new StringBuffer(orig);
      for(i=0; i<buf.length(); i++) {
	if(buf.charAt(i)=='\t'){
	  buf.setCharAt(i,' ');
	  buf.insert(i+1,spaces);
	}
      }//for
      return buf.toString();
    }//if
    else return orig;
  }

  /** Reduce names found in 'str' to contain RMI information -- for a neater display */
  //-----------------------------------
  private String reduceNames(String str) {
  //-----------------------------------
    
    StringTokenizer tok = new StringTokenizer(str, " \t\n\r", true);
    String curr;
    StringBuffer buf = new StringBuffer(str.length());

    while(tok.hasMoreTokens()) {
      curr = tok.nextToken();
      int begin = curr.indexOf("//");
      if(begin >= 0) {
	int end = curr.indexOf("/", begin + 2);
	  buf.append(curr.substring(0,begin));
	  if(end>=0)buf.append(curr.substring(end + 1));
      }
      else buf.append(curr);
    }
    return buf.toString();

  }//reduceNames



  /** Determines if the mouse being released at (x,y) has an effect on
    * the graphical appearance of this Link. */
  //----------------------------------------------
  public boolean mouseReleasedHasEffect(int x, int y) {
  //----------------------------------------------
    if (mousePressed) {
      mousePressed = false;
      return true;
    }
    return false;
  }//end mouseReleasedHasEffect


  /** Determines if the mouse being pressed at (x,y) has an effect on
    * the graphical appearance of this Link. */
  //----------------------------------------------
  public boolean mousePressedHasEffect(int x, int y) {
  //----------------------------------------------
    if (labelRect != null && labelRect.contains(x,y)) {
      mousePressed = true;
      return true;
    }
    return false;
  }//end mousePressedHasEffect



 /** Rotation and translation of points to a new 
     coordinate system. The new coordinate system is defined
     by two points p1,p2, and a (0,0) shift coordinate.
 
    @param p1 First point coordinate
    @param p2 Second point coordinate
    @param c  Systems (0,0) for the traslation
    @param vin Input vector of points.
    @param vout Ouput Traslated and rotated points.
  */

  private void rotateAndTranslate( Point p1, Point p2, Point c,
				   Point vin[], Point vout[] )  {
    //Rotation and shift to the real coordinate system.
    
    double x1,y1,theta;
    
    x1=(double) (p1.x -p2.x);
    y1=(double) (p1.y -p2.y);
    
    if (x1==0) theta=Math.PI  /2;
    else theta= Math.atan( y1/x1);
    if (x1 >0 ) theta+= Math.PI  ; //Correction  for the 2nd and 3rd quadrant
    
    for (int i=0; i< vin.length; i++)
    {
      int x,y;
      double xp,yp;
      
      xp=(double) vin[i].x;
      yp=(double) vin[i].y;
      
      x= (int)(xp * Math.cos(theta) - yp * Math.sin(theta) +c.x);
      y= (int)(xp * Math.sin(theta) + yp * Math.cos(theta) +c.y);
      
      vout[i]= new Point(x,y);
    }
  }//end rotateAndTranslate


  /** Get the bounding box for this graphical object */
  public Rectangle getBounds(){
    return bounds;
  }

  /** Get the Message this object represents */
  public Message getMessage() {
    return message;
  }

  /** Returns the full, formatted display string of the Message
    * this object represents
    */
  public String getContextString() {
    return message.getDisplayString(2);
  }
  
   /* Body of this method is empty. 
    * Needed to implement the methods of the DisplayEvent interface, 
    * but this is not needed here as it only has one index.
    * this method is the only one that is not used
    *@see kraft.monitor.display.DisplayEvent
    */
  /** @see kraft.monitor.agents.DisplayEvent */
  public void setCurrentIndex(int ix) {
    //do nothing
  }


  public void setCrossedOut(boolean state) {
    crossedOut = state;
  }

  /** @see kraft.monitor.agents.DisplayEvent */
  public void setContainerSize(Dimension s){
    parentSize = s;
  }


  private void setBounds(Rectangle r){
    bounds = r;
  }

  public void setBackground(Color c){
    background = c;
  }
  public Color getBackground(){
    return background;
  }
  public void setForeground(Color c){
    foreground = c;
  }
  public Color getForeground(){
    return foreground;
  }


  /** If this Link is associated with the Node oldy, replace it for the Node newby */
  //------------------------------------------
  public void switchNodes(Node oldy, Node newby) {
  //------------------------------------------
    Debug.print("Link("+label+").switchNodes entered: ");
    if(startNode.equals(oldy)) {
      startNode = newby;
      Debug.print("Switched start node\t");
      oldy.getPosition();
      newby.getPosition();
    }
    if(endNode.equals(oldy)) {
      Debug.print("Switched end node");
      endNode = newby; 
      oldy.getPosition();
      newby.getPosition();
   }
    Debug.println();
  }//end switchNodes
    

  /** @see kraft.monitor.agents.DisplayEvent */
  public void addDisplayMonitor(DisplayMonitor monitor){
    displayMonitor = monitor;
  }
  
  /** @see kraft.monitor.agents.DisplayEvent */
  public void tidyUp() {
    displayMonitor.setDisplayed(true);
  }

  /** Set the label of this Link to be 'str' */
  //------------------------------- 
  public void setLabel(String str) {
  //------------------------------- 
    label = str;
  }


  /** Re-calculate positions based on the positions of the two Nodes */
  //------------------------------
  private void updatePositions() {
  //------------------------------

    start    = startNode.getPosition();
    end      = endNode.getPosition();
    

    if (end.x >= start.x) {
      start.x += startNode.getRadius();
      end.x   -= endNode.getRadius();
    }else {
      start.x -= startNode.getRadius();
      end.x   += endNode.getRadius();
    }


    minX  = Math.min(start.x, end.x);
    minY  = Math.min(start.y, end.y);
    width = Math.abs(end.x - start.x);
    height= Math.abs(end.y - start.y);


    setBounds(new Rectangle(minX,minY,width,height));

    //Debug.println("Link: "+this+"\nhave updated positions:");
    //Debug.println("start node = "+startNode+", \nend node = "+endNode);
    

  }

  //----------------------- 
  public String toString() {
  //----------------------- 
    return "Link: [label = "+label+", start = ("+start+
      "), end = ("+end+")]";
  }

  //private Vector points;
  //private Enumeration enum;

}//end class Link














