package kraft.monitor.display;

import kraft.monitor.tools.Debug;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;


/** Provides the heart of the Monitor program.  Deals with actually graphically 
  * displaying visualisable events onto a graphical canvas.
  *<br><br>
  * E.g. an agent registering itself with the Monitor is represented as a red circle...<br>
  *      a message passed between two agents is represented as a blue arrow between to circles.
  *<br><br>
  * All these events are treated as much as possible in the same way by the MonitorCanvas,
  * as a <code> DisplayEvent </code> -- an event to be displayed.
  *<br><br>
  * The two DisplayEvents used in this canvas are:  <br>
  *<code>     Node    </code>: to represent a Monitorable agent <br>
  *<code>     Link    </code>: to represent a Message passed between two Monitorable agents <br>
  *<br>
  * Mouse events are passed from this canvas to the correct DisplayEvents (e.g. when a Node is clicked on)
  * to allow the DisplayEvent to show more information if appropriate.
  *<br><br>
  * To aid comprehension, the canvas is split into various areas -- represented by <code> Area </code>
  * objects which contain a number of <code> Node</code>s -- the correct area for a node is chosen 
  * according to the Monitorable type of the agent (e.g. <code> Monitorable.BROKER </code>).
  *
  *@author Ted Francis
  *@see kraft.monitor.display.MonitorGUI
  *@see kraft.monitor.display.DisplayEvent
  *@see kraft.monitor.display.DisplayEvent#mousePressedHasEffect
  *@see kraft.monitor.display.Node
  *@see kraft.monitor.display.Link
  *@see kraft.monitor.display.Area
  */

 /* NB: to the developer 
  * Various methods are provided which are called by the MonitorApplet to give information about the 
  * current visualisation step displayed.  The use of DisplayMonitors between the MonitorApplet
  * and the MonitorCanvas is used to ensure that a Display Event 'knows' what step it is at in the
  * visualisation -- a display monitor is passed to the canvas from the applet, the canvas assigns it a
  * step number, when the display event has been displayed, the displaymonitor is told this, so the 
  * MonitorApplet can ascertain when it is safe to continue allowing more information to be passed to it from
  * registered agents.
  */

//========================================
public class MonitorCanvas extends Canvas {
//========================================
 
  private Vector nodes; 
  private Vector displayEvents;
  private Area[] areas = null; 
  private int    eventPointer = 0, oldEventPointer = 0;
  private static final int FONT_SIZE = 12;

  //----------------------  
  public MonitorCanvas() {
  //----------------------  
    super();
    Debug.println("MonitorCanvas.<init>: constructor entered");
    nodes        = new Vector(5);
    displayEvents        = new Vector(10);
    eventPointer  = oldEventPointer = 0;
    this.setBackground(Color.white);
    this.addMouseListener(new CanvasEventHandler());
    this.setFont(new Font("SansSerif", Font.PLAIN, FONT_SIZE));
    
    repaint();
  }//end constructor


  /** A new Monitorable agent has registered with the monitor.
    * Add it to the canvas and handle the visualisation of it.<br><br>
    * (Creates a new Node object to deal with this Monitorable)
    *@param type the Monitorable type
    *@see kraft.monitor.display.Monitorable#BROKER
    */
  //--------------------------------------------------------  
  public void add(String name, Monitorable agent, int type) {
  //--------------------------------------------------------  
    
    Node newnode = new Node(name, agent, type);
    
    replaceDuplicates(name, newnode);

    nodes.addElement(newnode);
    areas[type].add(newnode);
    /* Debug.println("MonitorCanvas.add: Have added Node to set, total num = "+
		  nodes.size());
    */
    repaint();
  }//end add
  
  

  /** Removes the Node with the given name from the Nodes displayed
    * on the canvas. Used when an agent deRegisters.*/
  //-------------------------------
  public void remove(String name) {
  //-------------------------------

    Node b = getNode(name);
    if (b!=null) {
      nodes.removeElement(b);
      Area area = areas[b.getType()];
      area.remove(b);
      repaint();
    }
  }//end remove
  

  public String[] getMessages() {
    Vector v = new Vector(displayEvents.size());

    Enumeration e = displayEvents.elements();
    while(e.hasMoreElements()) {
      DisplayEvent link = (DisplayEvent)e.nextElement();
      if(link instanceof Link) {
	v.addElement(((Link)link).getMessage().toString());
      }
    }//end while
    
    String[] rtn = new String[v.size()];
    v.copyInto(rtn);
    
    return rtn;
  }//end getMessages
    


  
  /** Replace duplicate Nodes called 'name' with newnode
    * Maybe should indicate that it has died before removing it? */
  //----------------------------------------------------
  private void replaceDuplicates(String name, Node newnode) {
  //----------------------------------------------------
    Enumeration enum;

    Node old = getNode(name);
    if(old!=null) {
      enum = displayEvents.elements();
      while(enum.hasMoreElements()) {
	Object obj = enum.nextElement();
	if(obj instanceof Link)
	  ((Link)obj).switchNodes(old, newnode);
      }
      this.remove(name);
    }
  }//end replaceDuplicates




  /** Find the stored Node with the given name */
  //---------------------------------
  private Node getNode(String name) {
  //---------------------------------

    Enumeration e = nodes.elements();
    while (e.hasMoreElements()) {
      Node b = (Node) e.nextElement();
      if( b.getName().equals(name)){
	return b;
      }
    }
    return null; //if no matching Node is found
  }//end getNode
  
  
  //private Dimension oldsize;
  private boolean areas_initialised;



  /** Ensures the Areas are scaled correctly in relation to the size of
    * the MonitorCanvas */
  //----------------------
  private void setAreas() {
  //----------------------

    Dimension size = this.getSize();
    //Debug.println("Canvas size = "+size);

    if (!areas_initialised) {

      areas = new Area[4];
      
      areas[Monitorable.BROKER] = new Area(Monitorable.BROKER);
      areas[Monitorable.USER  ] = new Area(Monitorable.USER);
      areas[Monitorable.MIDDLE] = new Area(Monitorable.MIDDLE);
      areas[Monitorable.SERVER] = new Area(Monitorable.SERVER);

      areas_initialised = true;
    }//end if

    int BROKER_H, BROKER_W, NORMAL_H, NORMAL_W;
    
    double cnt = 0.00;

    for(int i=0; i<areas.length; i++) {
      if(areas[i].hasNodes() && i!=Monitorable.BROKER) //only check main areas 
	cnt += 1.00;                                   // not the broker area
    }

    if(cnt==0.00) NORMAL_W = 0;
    else NORMAL_W = (int)Math.ceil(size.width/cnt);

    if(areas[Monitorable.BROKER].hasNodes()) {
      BROKER_H = (int)Math.ceil(size.height/5.00);
      BROKER_W = size.width;
      NORMAL_H = size.height - BROKER_H; 
    }
    else{
      BROKER_H = 0;
      BROKER_W = 0;
      NORMAL_H = size.height;
    }

    int x = 0;
    int[] order = {Monitorable.USER, Monitorable.MIDDLE, Monitorable.SERVER};
    // the order to place areas other than the Broker area

    for(int i=0; i<order.length; i++) {
      Area a = areas[order[i]];

      if(a.hasNodes()) {
	a.setBounds(x, 0, NORMAL_W, NORMAL_H);
	x += NORMAL_W;
      }else
	a.setBounds(x, 0, 0, 0);
    }

    areas[Monitorable.BROKER].setBounds(0, NORMAL_H,
					BROKER_W, BROKER_H);

  }//end setAreas


  //---------------------------  
  public void update(Graphics g) {
  //---------------------------  
    paint(g);
  }


  /** Display the visualisation. Calls <code> paint(g) </code> on all the Areas
    * and the current DisplayEvent so that they display on the canvas.
    *@see kraft.monitor.display.Area
    *@see kraft.monitor.display.DisplayEvent
    */
  //---------------------------  
  public void paint(Graphics g) {
  //---------------------------  
    DisplayEvent next, prev;
    Area area;

    //super.paint(g);
    //System.out.println("MonitorCanvas.paint: entered");
    
    setAreas();
    
    if(areas[0]==null){
      g.setColor(Color.red);
      g.drawString("No display areas defined.", 20, 10);
      return;
    }
    if(!displayEvents.isEmpty() && oldEventPointer != eventPointer) {
      prev = (DisplayEvent)(displayEvents.elementAt(oldEventPointer));
      //prev.setCurrentIndex(eventPointer);
      prev.unpaint(g);
      //System.out.println("canvas: have unpainted :" +prev);
    }
    for(int i = 0; i < areas.length; i++) {
      area = areas[i];
      area.paint(g);
    }
    if(!displayEvents.isEmpty()) {
      Debug.println("\nMonitorCanvas.paint: about to paint elem at "+ 
			 eventPointer + "\n");
      next = (DisplayEvent)(displayEvents.elementAt(eventPointer));
      //next.setCurrentIndex(eventPointer);
      next.setContainerSize(this.getSize());
      next.paint(g);
      //System.out.println("canvas: have painted :" +next);
    }  

  }//end paint
  

  /** Go to the next step in the visualisation, if there is one.
    */
  //--------------------  
  public void nextStep(){
  //--------------------  
    gotoStep(eventPointer + 1);
  }//end nextStep
  

  /** Go to the previous step in the visualisation, if there is one.
    */
  //------------------------
  public void previousStep(){
  //------------------------
    gotoStep(eventPointer - 1);
  }//end previousStep


  /** Go to the step 'val' in the visualisation, if it exists.
    */
  //---------------------------
  public void gotoStep(int val) {
  //---------------------------
    oldEventPointer = eventPointer;

    if(val < 0 || displayEvents.size() <= 0)
      eventPointer = 0;
    else if (val >= displayEvents.size())
      eventPointer = displayEvents.size() -1;
    else eventPointer = val;

    if (eventPointer != oldEventPointer) {
      Enumeration e = displayEvents.elements();
      while(e.hasMoreElements()) {
	((DisplayEvent)e.nextElement()).setCurrentIndex(eventPointer);
      }
      repaint();
    }    
  }//end gotoStep



  //-------------------------------
  public Dimension getMinimumSize() {
  //-------------------------------
    return new Dimension(50, 50);
  }
  

  /** Clear all the display Events that are currently logged.
    * Effect will be to reduce the visualisation to just a picture
    * of the currently registered Monitorables.
    * The number of visualisation steps will then be zero.
    */
  //------------------------
  public void clearEvents() {
  //------------------------
    Enumeration enum = displayEvents.elements();

    //Stops the monitor looping forever waiting for the 
    // displayEvents to get displayed
    while(enum.hasMoreElements()) {
      ((DisplayEvent)enum.nextElement()).tidyUp();
    }
    displayEvents.removeAllElements();
    eventPointer = oldEventPointer = 0;
    this.repaint();
  }
  

  /** The current number of visualisation steps (display events)*/
  //-----------------------  
  public int totalSteps() {
  //-----------------------  
    return displayEvents.size();
  }
  

  /** The current visualisation step*/
  //----------------------
  public int currentStep() {
  //----------------------
    return eventPointer;
  }

  
  /** Get a current string that is relevant to the
    * current state of the visualisation.
    */
  //---------------------------------------
  public String getContextString() {
  //---------------------------------------
    
    if(displayEvents.isEmpty())
      return "";
    
    DisplayEvent event = (DisplayEvent)displayEvents.elementAt(eventPointer);
    return event.getContextString();
  }    


  /** Indicate a change in an agent's status.
    *@param name    the name of the Monitorable agent
    *@param comment the status 'comment' that should be used
    *@param dispmon the <code> setDisplayed(true) </code> method will be 
    *               called once this change has been visualised -- allows the
    *               caller to keep track of this happening.
    */
  //-------------------------------------------------------------
  public void statusChange(String name, String comment, 
			    DisplayMonitor dispmon) {
  //-------------------------------------------------------------
    Node b = getNode(name);

    displayEvents.addElement(b);
    int index = displayEvents.size()-1; //the index of the event

    dispmon.setIndex(index);
    b.addDisplayMonitor(dispmon);
    b.addComment(comment, index);    

    if(displayEvents.size() == 1)
      repaint();
  }


  /** Cross out the Link representing the given message.
    * Indicates that the Message was not correctly delivered.
    */
  //-------------------------------------------------------------
  public void crossOutMessage(String from, String to, Message msg) {
  //-------------------------------------------------------------

    DisplayEvent de;
    Link        rt;
    Debug.println("MonitorCanvas.crossOutMessage entered");

    for(int i = 0; i < displayEvents.size(); i++) {
	de = (DisplayEvent) displayEvents.elementAt(i);
	if(de instanceof Link) {
	  rt = (Link) de;
	  if(rt.getMessage().equals(msg)) {
	    rt.setCrossedOut(true);
	    if(i==eventPointer)
	      repaint();
	    return;
	  }
	}//end if(de...
    }//end for

  }//end crossOutMessage
    
	  
  
  /** Stores the given message in preparation for it to be displayed
    * on the canvas (as a Link between two Nodes).
    *
    *@param from The name of the Node this msg originates from
    *@param to   The name of the Node this msg is destined for
    *@param msg  The Message that is to be represented on screen
    *@param time The time the message was received 
    *@param dispmon To allow the calling object to keep track of
    *               the message and determine when it is displayed
    */

  //------------------------------------------------------------
  public void displayMessage(String from, String to, Message msg, 
			     Date time, DisplayMonitor dispmon) {
  //------------------------------------------------------------
    


    Node sender   = getNode(from);
    Node receiver = getNode(to);

    Link link  = new Link(sender, receiver, msg, time);
    
    displayEvents.addElement(link);
    dispmon.setIndex(displayEvents.size()-1);//index of its associated displayevent in the vector of displayevents.
    link.addDisplayMonitor(dispmon);

    if(displayEvents.size() == 1) {
      repaint();
    }    
  }//end displayMessage
  


  /** To handle and delegate any mouse events on this canvas.
    * Will find out which DisplayEvent the mouse event refers to
    * and inform it of the event.
    * The calls repaint to make these changes show.
    */
  class CanvasEventHandler extends MouseAdapter {

    public void mousePressed(MouseEvent e) {
      int mod = e.getModifiers();

      if( ! e.isPopupTrigger() &&
	  ((mod & MouseEvent.BUTTON1_MASK) == 0) && 
	  ((mod & MouseEvent.BUTTON2_MASK) == 0)) {


       	if( ! displayEvents.isEmpty() ) {
	  DisplayEvent de= (DisplayEvent)displayEvents.elementAt(eventPointer);
	  if(de.mousePressedHasEffect(e.getX(), e.getY())) {
	    repaint();
	    return; //don't affect the areas as displayEvents are always on top
	  }
	}
	for(int i =0; i < areas.length; i++) {
	  areas[i].setContainerSize(getSize());
	  areas[i].setGraphics(getGraphics());
	  if(areas[i].mousePressedHasEffect(e.getX(), e.getY()))
	     return; //don't check any more areas
	}
      }//end if
    }//end mousePressed
    

    public void mouseReleased(MouseEvent e) {
      int mod = e.getModifiers();

      if( ! e.isPopupTrigger() &&
	  ((mod & MouseEvent.BUTTON2_MASK) == 0) && 
	  ((mod & MouseEvent.BUTTON3_MASK) == 0)) {

	if(!displayEvents.isEmpty()) {
	  DisplayEvent de= (DisplayEvent)displayEvents.elementAt(eventPointer);
	  if(de.mouseReleasedHasEffect(e.getX(), e.getY())) {
	    repaint();
	    return;
	  }
	}
	for(int i =0; i < areas.length; i++) {
	  if(areas[i].mouseReleasedHasEffect(e.getX(), e.getY())) {
	    repaint();
	    return;
	  }
	}//for
      }//end if
    }//end mouseReleased

  }//end class CanvasEventHandler

}//end class MonitorCanvas














