package kraft.monitor.display;

import kraft.monitor.tools.Debug;

import java.awt.*;
import java.util.Vector;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Hashtable;


/** The graphical representation of a Monitorable agent in the visualisation tool.
  *<br><br>
  * Will draw the agent as a red circle with the agent's name nearby it -- the orientation of 
  * the label with respect to the circle can be changed.
  *<br><br>
  * This object will also store a set of status strings which are relevant at particular steps
  * in the visualisation -- calling setCurrentIndex will determine which status string is 
  * relevant at that point. <br><br>
  * If there is a relevant status string, the node's colour will be set to green. <br>
  * Mouse clicks on the object (via calls to <code> MousePressedHasEffect and 
  * MouseReleasedHasEffect </code>) 
  * will cause the currently relevant status string to be shown when the object is repainted.
  *
  *@author Ted Francis
  *@see #setCurrentIndex
  *@see kraft.monitor.display.Monitorable
  */

//========================================
public class Node implements DisplayEvent {
//========================================

  /** Constant for use with label orientation
    *@see #setLabelOrientation
    */
  public static final int
    NORTH     = 0,
    NORTHEAST = 1,
    EAST      = 2,
    SOUTHEAST = 3,
    SOUTH     = 4,
    SOUTHWEST = 5,
    WEST      = 6,
    NORTHWEST = 7;

  private static final int DIAMETER  = 16, RADIUS = DIAMETER/2;

  private String         monitorableName, label, startingComment, resource="", protocol="";
  private Monitorable    myObj;
  private Point          myPos;
  private int            myType, orient;
  private Rectangle      bounds, labelRect;

  private Hashtable      comments;
  private int            currentPos;
  private Vector         displayMonitors;
  private DisplayMonitor currentDisplayMonitor;           


  /** A new Node to represent a Monitorable object.
    *@param name the label for the Node
    *@param obj the monitorable object this Node represents grahically
    *@param type the Monitorable type of 'obj'
    *@see kraft.monitor.display.Monitorable
    */
  //--------------------------------------------
  public Node(String name, Monitorable obj, int type) {
  //--------------------------------------------
  
    this(name, obj);
    myType = type;
  }


  /** A new Node to represent a Monitorable object.
    *@param name the label for the Node
    *@param obj the monitorable object this Node represents grahically
    */
  //---------------------------------
  public Node(String name, Monitorable obj){
  //---------------------------------
    this(name, obj, new Point(5,5));
  }


  /** A new Node to represent a Monitorable object.
    *@param name the label for the Node
    *@param obj the monitorable object this Node represents grahically
    *@param pos the initial position of the Node
    *@see kraft.monitor.display.Monitorable
    */
  //---------------------------------------------
  Node(String name, Monitorable obj, Point pos) {
  //---------------------------------------------
    monitorableName = name;
    myObj           = obj;
    myPos           = pos;
    orient          = EAST;

    try{
      label            = obj.getDisplayName(1);
      startingComment  = obj.getDisplayName(2);
      resource = obj.getResourceType();
      protocol = obj.getResourceProtocol();

      //the one to default to if there is no currently relevant string
    }catch(Exception e){
      label = startingComment = name;
    }

    currentPos      = -1;
    comments   = new Hashtable(10); 
    // To contain different status msgs for the speech bubble


    labelRect = new Rectangle();

    displayMonitors       = new Vector(10);
    currentDisplayMonitor = null;
  }


  //------------------------------
  public void unpaint(Graphics g) {
  //------------------------------
    //Debug.println("Node["+label+"].UNpaint entered");
   
    drawTooltip(false, g);
  }


  /** Display this Node on the graphics context.
    * Set colour and display status string depending on the 
    * current index number we are at (ie. the visualisation step)
    *@see #setCurrentIndex
    */
  //----------------------------
  public void paint(Graphics g) {
  //----------------------------
    //Debug.println("Node["+label+"].paint entered");


    //Decide whether we are at a point where we have a comment string to
    // to show
    boolean haveComment = comments.containsKey(new Integer(currentPos));

    // Uncomment the line below to make the resource of the agent be drawn,
    // if it has one:-

    // drawResource(mousePressed,g);
    
    if(haveComment)
      g.setColor(Color.green);
    else
      g.setColor(Color.red);

    g.fillOval(myPos.x-RADIUS, myPos.y-RADIUS, DIAMETER, DIAMETER);
    
    Point pos = getLabelPos(g.getFontMetrics());
    
    g.setColor(Color. black);
    g.drawString(label, pos.x, pos.y);


    if(mousePressed) {
      drawTooltip(true, g);
    }

    if(haveComment) 
      finishCurrentDisplayMonitor(); //as we have now displayed it.
    
  }//end paint


  private boolean justbeenpressed = false;


  /**Draws the resource of this agent. Called from the paint() method, although
    *is commented out at the moment
    */
  private void drawResource(boolean mousePressed, Graphics g) {
    int w, h, line_w = DIAMETER *2, box_w;
    FontMetrics fm;
    
    boolean have_resource, have_protocol;
    have_resource = (!(resource==null || resource.equals("")));
    have_protocol = (!(protocol==null || protocol.equals("")));

    if( !justbeenpressed && have_resource ) {
      System.out.println();
      System.out.println(label+": have resource '"+resource);

      setLabelOrientation(SOUTH);
      fm = g.getFontMetrics();
      w  = box_w  = fm.stringWidth(resource) + 4;
    
      h = fm.getMaxAscent() + 4;

      if(have_protocol && fm.stringWidth(protocol)>line_w) 
	line_w = fm.stringWidth(protocol)+4;

      w += line_w;
      
      System.out.println("w = " +w);
      System.out.println("h = " +h);
      System.out.println("line_w = " +line_w);
      System.out.println("box_w = " + box_w);


      myPos.x = myPos.x - w/2;  //adjust pos of circle

      	
      System.out.println("myPos changed to = " +myPos);

      if(!mousePressed) {
	//now do the resource drawing
	g.setColor(Color.lightGray);
	g.fillRect(myPos.x+RADIUS+line_w+2, myPos.y-h/2+1, box_w+1, h+1);
	g.setColor(Color.black);
	g.drawRect(myPos.x+RADIUS+line_w+1, myPos.y-h/2, box_w, h);
	g.drawString(resource,myPos.x+RADIUS+line_w+3, myPos.y + h/2 -2);
	if(have_protocol) {
	  g.setColor(Color.darkGray);
	  g.drawString(protocol,myPos.x+RADIUS+3, myPos.y-2);
	}
	g.setColor(Color.black);
	g.drawLine(myPos.x+RADIUS, myPos.y, myPos.x+RADIUS+line_w+1, myPos.y);
	System.out.println("done drawing resource");
      }
    }//end if(have_resource)
    else
      System.out.println(label+": have no resource");

    if(mousePressed && !justbeenpressed) justbeenpressed = true;

    else if(!mousePressed && justbeenpressed) justbeenpressed = false;
  }//end drawResource



  private String stringToShow;
  


  /** Returns the current status message of this
    * Node.
    */
  public String getContextString() {
    if(stringToShow.equals(startingComment))
      return "";
    else
    return (label + "\n\n" + stringToShow);
  }

  /** Draw the currently relevant status string as a 'pop-up tooltip' 
    * onto the graphics context if drawMe==true, else ensure it is 
    * fully undrawn from the graphics context. */
  //-----------------------------------------------------------------
  private void drawTooltip(boolean drawMe, Graphics gc) {
  //-----------------------------------------------------------------
    int             x, y, w, h, maxW, maxH, space;
    FontMetrics     fm;
    Vector          strings;
    StringTokenizer tok;
    boolean         contextOn;


    stringToShow = (String)comments.get(new Integer(currentPos));

    contextOn = true;
    //System.out.println("\n"+label+".drawTooltip: comment at "+currentPos);


    if(stringToShow==null) {
      stringToShow = startingComment;
      contextOn  = false;
    }
    stringToShow = replaceTabs(stringToShow, 5);
    Debug.println(""+label+".drawTooltip: stringToShow = "+stringToShow+"\n");
    
    fm    = gc.getFontMetrics(); 
    space = 5; //fm.getHeight();

    tok  = new StringTokenizer(stringToShow, "\n\r");
      
    maxH = parentSize.height - 1;
    maxW = parentSize.width  - 1;

    strings = new Vector(tok.countTokens());
    
    w = 0;
    while (tok.hasMoreTokens())
    {
      String curr = tok.nextToken();
      int    strW = fm.stringWidth(curr);
      
      if(strW + space*2 <= maxW)
      {
	strings.addElement(curr);
	w = Math.max(w, strW + space*2);
      }//end if

      else //string is too wide
      { //need to chop up the string into 'num' parts
	int num = (int)Math.ceil((double)strW/ (maxW-space*2));  
	//each part will be 'len' chars long
	int len = curr.length()/num;
	
	for(int pos = 0; pos < curr.length() ; pos+=len)
	{
	  int end; String tmp;
	  end = (pos+len>=curr.length()) ? curr.length()-1:pos+len;
	  tmp = curr.substring(pos, end);
	  strings.addElement(tmp);
	  w   = Math.max(w, fm.stringWidth(tmp)+space*2);
	}
      }//end else
    }//end while

    h = strings.size() * fm.getHeight();
    
    x = Math.max(0, myPos.x - w/2);
    y = Math.max(0, myPos.y - h/2);
    
    if(y+h > maxH) { 
      y = Math.max(0, y - (y+h-maxH));
    }
    if(x+w > maxW) {
	x = Math.max(0, x - (x+w-maxW));
    }

    
    if(drawMe)
      gc.setColor(Color.getHSBColor(((float)258)/360,(float)0.14,(float)0.96));
    //a mauve colour.

    //gc.setColor(Color.getHSBColor(((float)267)/360, (float)0.2, (float)0.9));
    //a violet colour (yuk!)

    else
      gc.setColor(Color.white);

    gc.fillRect(x,y,w,h);
    if(drawMe) {
	gc.setColor(Color.black);
    }

    gc.drawRect(x,y,w,h);

    x += space; 
    y += fm.getMaxAscent();
    
    if(drawMe) {
      gc.setColor(Color.black);
      Enumeration e = strings.elements();
      while(e.hasMoreElements()) {
	gc.drawString((String)e.nextElement(), x, y);
	y += fm.getHeight();
      }
    }

  }//end drawTooltip


  private String replaceTabs(String orig, int num_spaces){
    if(orig.indexOf('\t')>=0) { //there is a tab
      String spaces ="";
      int i;
      for(i=1;i<num_spaces;i++)
	spaces += " ";
      StringBuffer buf = new StringBuffer(orig);
      for(i=0; i<buf.length(); i++) {
	if(buf.charAt(i)=='\t'){
	  buf.setCharAt(i,' ');
	  buf.insert(i+1,spaces);
	}
      }//for
      return buf.toString();
    }//if
    else return orig;
  }



  /** Set the current index, so this object knows which status string is 
    * currently relevant at this visualisation step -- if there is one.
    */
  //---------------------------------
  public void setCurrentIndex(int ix) {
  //---------------------------------
    currentPos = ix;
    //System.out.println(label+"currentPos set to " + currentPos);
  }


  /** Give the Node a status string for the visualisation step 'index'.
    * Will be displayed when <code> setCurrentIndex(index) </code> is called */
  //---------------------------------------------
  public void addComment(String comment, int index) {
  //---------------------------------------------
    StringTokenizer tok = new StringTokenizer(comment, " \t\n", true);
    StringBuffer buf    = new StringBuffer();

    while(tok.hasMoreTokens()) {
      buf.append(simpleName(tok.nextToken()));
    }

    comments.put(new Integer(index), buf.toString());

    Debug.println("\n"+label+"addComment: have added comment with key: "+index+", comment is \""+comment+"\"\n");

  }

  /** Get the host name from the given RMI name. */
  //-----------------------------------------
  public static String hostName(String rminame) {
  //-----------------------------------------

    String tmp = "";
    int begin = rminame.indexOf("//");
    if(begin>=0){
      int end = rminame.indexOf(":", begin + 2);
      if(end<0) end   = rminame.indexOf("/", begin + 2);
      if(end<0) end   = begin + 3;
      tmp = rminame.substring(begin+2, end);
    }

    try {
      return java.net.InetAddress.getByName(tmp).getHostName();
    }
    catch(java.net.UnknownHostException e) {
      return tmp;
    }
  }//end hostName

  
  /** Get the simple 'agent' name from the given RMI name */
  //--------------------------------------------
  public static String simpleName(String rminame) {
  //--------------------------------------------
    String tmp = rminame;

    int begin = rminame.indexOf("//");
    if(begin>=0){
      int end = rminame.indexOf("/", begin + 2);
      if(end>=0) tmp = rminame.substring(end+1);
      else       tmp = rminame.substring(begin+2);
    }
    if(tmp.endsWith("_Monitorable"))
      tmp = tmp.substring(0, tmp.indexOf("_Monitorable"));
    
    return tmp;
  }//end simpleName  



  private Dimension parentSize;

  /** @see kraft.monitor.display.DisplayEvent */
  //--------------------------------------
  public void setContainerSize(Dimension d) {
  //---------------------------------------
    parentSize = d;
  }


  /** @see kraft.monitor.display.DisplayEvent */
  //-------------------
  public void tidyUp() {
  //------------------
    Enumeration e = displayMonitors.elements();
    while(e.hasMoreElements()) {
      ((DisplayMonitor)e.nextElement()).setDisplayed(true);
    }
    displayMonitors.removeAllElements();
    comments.clear();
    currentPos = -1;
  }//end setDisplayed



  /** @see kraft.monitor.display.DisplayEvent */
  //---------------------------------------------
  public void addDisplayMonitor(DisplayMonitor dm) {
  //---------------------------------------------
    displayMonitors.addElement(dm);
    if(currentDisplayMonitor == null) {
      currentDisplayMonitor = dm;
      //currentPos            = dm.getIndex();
    }
    Debug.println("\n"+label+".addDisplayMonitor: added display monitor: "+dm+"\n\tcurrentDispMon = " + currentDisplayMonitor + "\n\tcurrentPos = "+currentPos+"\n");
    
  }

  /** Tell the current display monitor that its display event has been displayed.*/
  //------------------------------------------
  private void finishCurrentDisplayMonitor() {
  //------------------------------------------

    if(currentDisplayMonitor != null) {
      currentDisplayMonitor.setDisplayed(true); //ensure nothing continues
                                                // waiting on this DispMonitor
    Debug.println("\n"+label+".finishCurrent...: removing "+currentDisplayMonitor+ " from vector.");

      displayMonitors.removeElement(currentDisplayMonitor);
      if(displayMonitors.isEmpty()) {
	currentDisplayMonitor = null;
      }
      else {
	currentDisplayMonitor = (DisplayMonitor)displayMonitors.firstElement();
      }
      Debug.println("\n"+label+".finishCurrent...: currentDisplayMonitor is now " + currentDisplayMonitor + " currentPos = "+currentPos+"\n");
  
  }

  }//end finishCurrentDisplayMonitor

  

  /** Return whether this object contains the point (x,y) */
  //-----------------------------------
  public boolean contains(int x, int y) {
  //-----------------------------------

    boolean val = (x >= myPos.x - RADIUS && 
		   x <= myPos.x + RADIUS &&
		   y >= myPos.y - RADIUS &&
		   y <= myPos.y + RADIUS);
    return val;
  }


  private boolean mousePressed = false;


  /** If a mouse press at (x,y) has an effect on this Node-- i.e. there is a relevant 
    * status string that can be popped up 
    */
  //---------------------------------------------
  public boolean mousePressedHasEffect(int x, int y) {
  //---------------------------------------------
    if(!mousePressed && contains(x,y)) {
      mousePressed = true;
      return true;
    }
    return false;
  }


  /** If a mouse releas at (x,y) has an effect on this Node-- i.e. there is a relevant 
    * status string that can be popped down 
    */
  //---------------------------------------------
  public boolean mouseReleasedHasEffect(int x, int y) {
  //---------------------------------------------
    if(mousePressed) {
      mousePressed = false;
      return true;
    }
    return false;
  }


//  public void mousePressed(boolean state, int x, int y) {
//     mousePressed = state;
//   }



  /** Ascertain the label position depending on the orientation */
  //---------------------------------------
  private Point getLabelPos(FontMetrics fm) {
  //---------------------------------------
    int x = 0, y = 0;

    if(orient < NORTH || orient > NORTHWEST)
      orient = EAST;

    // work out x-coord, 
    if(orient == NORTH || orient == SOUTH)
      x =  myPos.x - fm.stringWidth(label)/2;
    if(orient == EAST || orient == NORTHEAST || orient == SOUTHEAST)
      x = myPos.x + RADIUS + 3;
    if(orient == WEST || orient == NORTHWEST || orient == SOUTHWEST)
      x = myPos.x - RADIUS - fm.stringWidth(label) - 3;

    //work out y-coord
    if(orient == NORTH || orient == NORTHEAST || orient == NORTHWEST)
      y = myPos.y - RADIUS -fm.getHeight()/2 - 3; 
    if(orient == SOUTH || orient == SOUTHEAST || orient == SOUTHWEST)
       y = myPos.y + RADIUS +fm.getHeight()/2 + 3; 
    if(orient == EAST || orient == WEST)
      y = myPos.y + fm.getHeight()/2;

    labelRect.setLocation(x,y);
    labelRect.setSize(fm.stringWidth(label), y - fm.getMaxAscent());

    return new Point(x,y);

  }//end getLabelPos


  //-------------------------------
  // Accessor Methods
  //-------------------------------

  public Point getPosition() {
      return new Point(myPos);
  }

  
  public String getName(){
    return monitorableName;
  }


  /** Returns the type of Monitorable object this Node represents
    *@see kraft.monitor.display.Monitorable#BROKER
    */
  public int getType() {
    return myType;
  }

  public int getRadius() {
    return RADIUS;
  }



  //--------------------------------
  // Mutator Methods
  //---------------------------------

  
  public void setBounds(Rectangle r) {
    bounds = r;
  }

  public void setPosition(Point newpos){
      myPos = newpos;
  }

  public void setPosition(int x, int y){
    this.setPosition(new Point(x,y));
  }

  public void setLabel(String str){
    label = str;
  }


  /** Set the orientation of the label with respect to the Node.
    *@see #NORTH
    */
  public void setLabelOrientation(int ori){
    orient= ori;
  }


  public String toString(){
    return "Node: [label = "+label+", pos = "+getPosition()+", labelorient = "+orient+
      "]";
  }

}//end class Node














