%
% socket library for QUINTUS Prolog
%
%

:- use_module('util/prolog_detect').
:- prolog_detect:prolog(Prolog),
	(	(Prolog=sicstus,
		use_module(library(lists))
		);
		(Prolog=quintus,
		use_module(library(basics)),
		use_module(library(lists)),
		use_module(library(socket_io))
		)
	).


%----------------------
% Prolog to C interface
%----------------------

:- dynamic foreign/2, foreign_file/2.

%
% C function to Prolog predicate mappings
%
% NOTE: The Qunitus interface uses an integer to pass back the result of the C function
%       0 means ok. no-zero means error.
%       Every C function pass back the result as the last parameter.
%

foreign(c_socket_create_I,c_socket_create_I(-integer,[-integer])).
foreign(c_socket_create_U,c_socket_create_U(-integer,[-integer])).
foreign(c_socket_destroy,c_socket_destroy(+integer,[-integer])).
foreign(c_socket_bind_I,c_socket_bind_I(+integer,+integer,-integer,[-integer])).
foreign(c_socket_bind_U,c_socket_bind_U(+integer,+string,[-integer])).
foreign(c_socket_listen,c_socket_listen(+integer,+integer,[-integer])).
foreign(c_socket_accept,c_socket_accept(+integer,-integer,[-integer])).
foreign(c_socket_connect_I,c_socket_connect_I(+integer,+string,+integer,[-integer])).
foreign(c_socket_connect_U,c_socket_connect_U(+integer,+string,[-integer])).
foreign(c_socket_select,c_socket_select(+term,+term,+term,+integer,+integer,-term,-term,-term,[-integer])).
foreign(c_hostname,c_hostname(-string,[-integer])).

%
% C functions to be found in object file sockets.o
%

foreign_file('sparc_quintus_sockets.o',	[
				c_socket_create_I,
				c_socket_create_U,
				c_socket_destroy,
				c_socket_bind_I,
				c_socket_bind_U,
				c_socket_listen,
				c_socket_accept,
				c_socket_connect_I,
				c_socket_connect_U,
				c_socket_select,
				c_hostname
				]).

%
% finds out what libraries are needed by 'sockets.o'
% and loads in 'sockets.o'
%
% NOTE: Qunitus does not have to 'reqlib' as in Sicstus.
%       It needs the C 'socket' library anyway.
%

:- load_foreign_files('sparc_quintus_sockets.o',['-lsocket']).




%---------------------------------------------
% create a socket in AF_INET or AF_UNIX domain
%
% socket_create(+Domain,-Socket)
%
% Domain=	'AF_INET' for Internet domain socket
%		'AF_UNIX' for UNIX domain socket
% Socket=	socket to be returned
%---------------------------------------------

socket_create(Domain,Socket)	:-
	nonvar(Domain),
	socket_domain(Domain),
	var(Socket),
	socket_create1(Domain,Socket).

socket_domain('AF_INET').
socket_domain('AF_UNIX').

%
% create a socket in the Internet domain
%
socket_create1('AF_INET',Socket)	:-
	c_socket_create_I(Socket,0).

%
% create a socket in the UNIX domain
%

socket_create1('AF_UNIX',Socket)	:-
	c_socket_create_U(Socket,0).




%-----------------
% destroy a socket
% socket_destroy(+Socket)
%
% Socket=socket to be destroyed
%-----------------

socket_destroy(Socket)	:-
	nonvar(Socket),
	integer(Socket),
	c_socket_destroy(Socket,0).





%-------------------------------
% binds a socket with an address
% socket_bind(+Socket,+Address)
%
% Socket=	socket to be binded (created by socket_create/2)
% Address=	address in the form 'AF_INET'(hostname,port)
% or		'AF_UNIX'(pathname)
%-------------------------------

socket_bind(Socket,Address)	:-
	integer(Socket),
	nonvar(Address),
	socket_bind1(Socket,Address).

%
% Internet socket, port is given
%
socket_bind1(Socket,'AF_INET'(Port))	:-
	nonvar(Port),
	!,
	c_socket_bind_I(Socket,Port,_,0).		%ignore returned port

%
% Internet socket, port is not given
%
socket_bind1(Socket,'AF_INET'(Port))	:-
	c_socket_bind_I(Socket,0,Port,0).		%take returned port

%
% UNIX socket
%
socket_bind1(Socket,'AF_UNIX'(Path))	:-
	atom(Path),
	c_socket_bind_U(Socket,Path,0).




%----------------------------------------
% listen to a socket and set queue length
% socket_listen(+Socket,+Length)
%
% Socket=	socket to listen (created by socket_create/2)
% Length=	integer specifying the queue length
%----------------------------------------
socket_listen(Socket,Length)	:-
	integer(Socket),
	integer(Length),
	!,
	c_socket_listen(Socket,Length,0).




%------------------------------
% accept an incoming connection
% socket_accept(+Socket,-Income,-Read,Write)
%
% Socket=	socket for accepting incoming connection
% Income=	returned file-descriptor of incoming connection
% Read=		read stream for incoming connection
% Write=	write stream for incoming connection
%------------------------------
socket_accept(Socket,Income,Read,Write)	:-
	integer(Socket),
	var(Income),
	var(Read),
	var(Write),
	!,
	c_socket_accept(Socket,Income,0),		%returns an incoming file-descriptor
	socket_io_open_input(Income,Read_code),		%call Quintus library for buffered input stream
	socket_io_open_output(Income,Write_code),	%output stream
	stream_code(Read,Read_code),			%convert SP_stream pointers to Prolog stream
	stream_code(Write,Write_code).




%---------------------------
% connect to a remote socket
% socket_connect(+Socket,+Address,+Read,+Write)
%
% Socket=	socket for making connection (created by socket_create/2)
% Address=	target address for connection in the form 'AF_INET'(host,port)
% or		'AF_UNIX'(pathname) for the UNIX domain
% Read=		read stream to be returned for connection
% Write=	write stream to be returned
%---------------------------
socket_connect(Socket,Address,Read,Write)	:-
	integer(Socket),
	var(Read),
	var(Write),
	socket_connect1(Socket,Address,Read,Write).

socket_connect1(Socket,'AF_INET'(Host,Port),Read,Write)	:-
	nonvar(Host),
	nonvar(Port),
	c_socket_connect_I(Socket,Host,Port,0),
	socket_io_open_input(Socket,Read_code),		%call Quintus library for buffered input stream
	socket_io_open_output(Socket,Write_code),	%output stream
	stream_code(Read,Read_code),			%convert SP_stream pointers to Prolog stream
	stream_code(Write,Write_code).

socket_connect1(Socket,'AF_UNIX'(Path),Read,Write)	:-
	nonvar(Path),
	c_socket_connect_U(Socket,Path,0),
	socket_io_open_input(Socket,Read_code),		%call Quintus library for buffered input stream
	socket_io_open_output(Socket,Write_code),	%output stream
	stream_code(Read,Read_code),			%convert SP_stream pointers to Prolog stream
	stream_code(Write,Write_code).




%----------------------------------------------------------------------
% select from lists of file descriptors
% socket_select(+Reads,+Writies,+Excepts,+Timeout,-Read,-Write,-Except)
%
% Reads=	list of file-descriptors for testing read-ready
% Writes=	list of file-descriptors for testing write-ready
% Excepts=	list of file-descriptors for testing exceptions
% Timeout=	time-out value for selection
%		either 0 for no timeout (blocking)
%		or time(Second,Micro-second) for timeout
% Read=		returned list of file-descriptors ready for reading
% Write=	returned list of file-descriptors ready for writing
% Except=	returned list of file-descriptors with exception
%----------------------------------------------------------------------

socket_select(Reads,Writes,Excepts,0,R,W,E)	:-
	socket_select1(Reads,Writes,Excepts,0,0,R,W,E).

socket_select(Reads,Writes,Excepts,time(Sec,Usec),R,W,E)	:-
	socket_select1(Reads,Writes,Excepts,Sec,Usec,R,W,E).

socket_select1(Reads,Writes,Excepts,Sec,Usec,R,W,E)	:-
	is_list(Reads),
	is_list(Writes),
	is_list(Excepts),
	integer(Sec),
	integer(Usec),
	!,
	c_socket_select(Reads,Writes,Excepts,Sec,Usec,R,W,E,0).





%------------------------------
% get the current hostname
% hostname(-Host)
%
% Host=	hostname to be returned
%------------------------------
hostname(Host)	:-
	c_hostname(Host,0).

